from collections import defaultdict
from typing import Dict, List

from kokex.core.parser import DocumentParser


def keywords(docs: List[str]) -> Dict[str, int]:
    """
    문서 목록을 받아서 포함된 키워드를 리턴합니다

    :param docs: 문서 목록
    :return: 키워드와 빈도가 담긴 딕셔너리
    """
    result = defaultdict(int)
    parser = DocumentParser()

    for doc in docs:
        parser.parse(document=doc)

        for noun in parser.nouns():
            result[noun] += 1

    return result


def sentences(doc: str) -> List[str]:
    """
    문서를 입력 받아 문장으로 분리한 리스트를 리턴합니다

    :param doc: 문서
    :return: 문장으로 분리한 리스트
    """
    parser = DocumentParser()
    parser.parse(document=doc)

    return parser.sentences()


def parse(doc: str, debug: bool = True):
    """
    문서를 입력받아서 파싱된 결과를 문자열로 리턴합니다

    :param doc: 입력 문서
    :param debug: true 일 경우 문서위계, 5언 7성분 9품사 정보를 함께 출력 (기본값 true)
    :return: 출력을 위해 들여쓰기가 된 문자열
    """
    parser = DocumentParser()
    parser.parse(doc)

    return parser.printable_tree(debug=debug)



import html
import re


# preprocessing 함수
def preproc(document):
    document = _html_unescape(document)
    document = _replace_html_tag(document)
    document = _replace_url_link(document)
    document = _replace_email(document)
    document = _replace_mention(document)
    document = _replace_dots(document)
    document = _replace_kkks(document)
    document = _replace_unicode_char(document)
    return document


def _html_unescape(text):
    return html.unescape(text)


def _replace_html_tag(text):
    re_pattern = r"</?[a-zA-Z][a-zA-Z0-9]*>"
    return re.sub(pattern=re_pattern, repl="", string=text)


def _replace_url_link(text):
    re_pattern = r"(http|ftp|https)://(?:[-\w.]|(?:[%/\w]))+(\?.+?=.+?(&.+?=.+?)*\b)?"
    return re.sub(pattern=re_pattern, repl="", string=text)


def _replace_email(text):
    re_pattern = r"[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+"
    return re.sub(pattern=re_pattern, repl="", string=text)


def _replace_mention(text):
    re_pattern = r"@[a-zA-Z0-9_.+-]+"
    return re.sub(pattern=re_pattern, repl="", string=text)


def _replace_unicode_char(text):
    u_chars = "".join(
        re.findall(pattern=r"(\\u[0-9a-zA-Z]{4}|\\x[0-9a-zA-Z]{2})", string=repr(text))
    )
    return re.sub(pattern=fr"[{u_chars}]", repl=" ", string=text) if u_chars else text


def _replace_dots(text):
    re_pattern = r"((·{2,})|(\.{2,}))"
    return re.sub(pattern=re_pattern, repl="… ", string=text)


def _replace_kkks(text):
    re_pattern = r"(ㅋ{3,})"
    return re.sub(pattern=re_pattern, repl="ㅋㅋㅋ ", string=text)



from fastapi.testclient import TestClient

import kokex
from kokex.server import server

client = TestClient(server.app)


def test_keywords_0001():
    check_results(
        input_documents=[
            "첫 번째 문서입니다. 여러 문장을 포함할 수 있습니다.",
            "두 번째 문서입니다. 여러 문서를 포함할 수 있습니다.",
        ],
        expected_results={"번째": 2, "문서": 3, "문장": 1, "포함": 2},
    )


def test_keywords_0002():
    check_results(
        input_documents=[
            "새로운 테스트 문장을 일련번호와 함께 메소드로 추가합니다.",
        ],
        expected_results={
            "테스트": 1,
            "문장": 1,
            "일련번호": 1,
            "일련": 1,
            "번호": 1,
            "메소드": 1,
            "추가": 1,
        },
    )


def check_results(input_documents, expected_results):
    keywords = kokex.keywords(input_documents)
    assert keywords == expected_results

    response = client.post("/keywords", json={"docs": input_documents})
    assert response.status_code == 200
    assert response.json() == expected_results



import re

import networkx as nx


class NodeData:
    def __init__(
        self,
        node_id,
        node_type,
        parent_node_id=None,
        word_tag=None,
        sentence_tag=None,
        semantic_tag=None,
        org_txt_form=None,
        pos_txt_form=None,
    ):

        self.__node_id = node_id
        self.__node_type = node_type
        self.__parent_node_id = parent_node_id

        self.__word_tag = word_tag
        self.__sentence_tag = sentence_tag
        self.__semantic_tag = semantic_tag

        self.__org_txt_form = org_txt_form
        self.__pos_txt_form = pos_txt_form

    @property
    def node_id(self):
        return self.__node_id

    @node_id.setter
    def node_id(self, value):
        self.__node_id = value

    @property
    def node_type(self):
        return self.__node_type

    @node_type.setter
    def node_type(self, value):
        self.__node_type = value

    @property
    def parent_node_id(self):
        return self.__parent_node_id

    @parent_node_id.setter
    def parent_node_id(self, value):
        self.__parent_node_id = value

    @property
    def word_tag(self):
        return self.__word_tag

    @word_tag.setter
    def word_tag(self, value):
        self.__word_tag = value

    @property
    def sentence_tag(self):
        return self.__sentence_tag

    @sentence_tag.setter
    def sentence_tag(self, value):
        self.__sentence_tag = value

    @property
    def semantic_tag(self):
        return self.__semantic_tag

    @semantic_tag.setter
    def semantic_tag(self, value):
        self.__semantic_tag = value

    @property
    def org_txt_form(self):
        return self.__org_txt_form

    @org_txt_form.setter
    def org_txt_form(self, value):
        self.__org_txt_form = value

    @property
    def pos_txt_form(self):
        return self.__pos_txt_form

    @pos_txt_form.setter
    def pos_txt_form(self, value):
        self.__pos_txt_form = value

    def get_last_pos_tag(self):
        p = re.compile(r"(?<=[/+])[A-Z]+$")
        return p.findall(self.pos_txt_form)[0]

    def get_first_pos_tag(self):
        p = re.compile(r"(?<=/)[A-Z]+")
        return p.findall(self.pos_txt_form)[0]


class ParseTree:
    ID_ROOT = "root"

    def __init__(self):
        self.g = nx.DiGraph()
        self.root = None

    def clear(self):
        self.g.clear()
        self.root = None

    def add_node(self, node_id: str, node_data: NodeData):
        # 노드 추가
        self.g.add_node(node_id, data=node_data)

        # 루트 노드가 아니라면 부모 노드와 에지 추가
        if node_data.parent_node_id:
            self.g.add_edge(node_data.parent_node_id, node_id)

        # 추가한 노드에 대한 후처리: 5언 7성분을 태깅
        if node_data.node_type in ["단어", "구"]:
            node_data.word_tag = self._compute_word_tag(node_data.get_last_pos_tag())
        if node_data.node_type in ["단어", "구", "절"]:
            node_data.sentence_tag = self._compute_sentence_tag(
                node_data.get_last_pos_tag()
            )

    def get_node_data_by_id(self, node_id: str):
        return self.g.nodes[node_id]["data"]

    def get_children_node_ids(self, parent_node_id: str):
        return sorted(self.g.adj[parent_node_id])

    def remove_node(self, node_id: str):
        self.g.remove_node(node_id)

    def filter_nodes(self, func):
        return filter(func, self.g.nodes)

    def is_leaf(self, node_id: str):
        return len(self.g.adj[node_id]) == 0

    def printable_subtree(self, sub_root_node_id, debug=True):
        node_depth = len(sub_root_node_id.split("_")) - 1
        node_data = self.g.nodes[sub_root_node_id]["data"]
        printable = ("\t" * node_depth) + "[" + node_data.node_id + "] "
        if debug:
            printable += (
                "["
                + node_data.node_type
                + "] "
                + "["
                + (node_data.word_tag if node_data.word_tag else "")
                + "] "
                + "["
                + (node_data.sentence_tag if node_data.sentence_tag else "")
                + "] "
                + node_data.pos_txt_form
            )
        else:
            printable += node_data.org_txt_form
        printable += "\n"

        for child_id in self.get_children_node_ids(sub_root_node_id):
            printable += self.printable_subtree(child_id, debug=debug)

        return printable

    # 노드 태그 관련 유틸리티함수 시작
    @staticmethod
    def _compute_word_tag(last_pos_tag: str):
        # 5언 계산
        if last_pos_tag[0] == "N" or last_pos_tag in ["ETN", "XPN", "XSN", "SH"]:
            word_tag = "체언"
        elif last_pos_tag[0] == "V" or last_pos_tag in ["EP", "EF", "EC", "XSV", "XSA"]:
            word_tag = "용언"
        elif last_pos_tag[0] == "M" or last_pos_tag in ["ETM"]:
            word_tag = "수식언"
        elif last_pos_tag[0] == "J" or last_pos_tag in []:
            word_tag = "관계언"
        else:
            word_tag = "독립언"

        return word_tag

    @staticmethod
    def _compute_sentence_tag(last_pos_tag: str):
        # 7성분 계산
        if last_pos_tag in ["JKS"]:
            sentence_tag = "주어"
        elif last_pos_tag in ["EP", "EF", "EC", "ETN"]:
            sentence_tag = "서술어"
        elif last_pos_tag in ["JKO"]:
            sentence_tag = "목적어"
        elif last_pos_tag in ["JX", "JKC"]:
            sentence_tag = "보어"
        elif last_pos_tag in ["JKB", "MAG"]:
            sentence_tag = "부사어"
        elif last_pos_tag in ["ETM", "JKG", "MM"]:
            sentence_tag = "관형어"
        else:
            sentence_tag = "독립어"

        return sentence_tag



from fastapi.testclient import TestClient

import kokex
from kokex.server import server

client = TestClient(server.app)


def test_sentences_0001():
    check_results(
        input_document="첫 번째 문서입니다. 여러 문장을 포함할 수 있습니다.",
        expected_results=["첫 번째 문서입니다.", "여러 문장을 포함할 수 있습니다."],
    )


def check_results(input_document, expected_results):
    sentences = kokex.sentences(input_document)
    assert sentences == expected_results

    response = client.post("/sentences", json={"doc": input_document})
    assert response.status_code == 200
    assert response.json() == expected_results



import re

from konlpy.tag import Mecab

from .preproc import preproc
from .tree import NodeData, ParseTree


class DocumentParser:
    def __init__(self):
        self.document = ""
        self.morphs = []
        self.mecab = Mecab()

        # tree initialization
        self.tree = ParseTree()

    def parse(self, document):
        # preprocessing
        self.document = preproc(document)
        self.morphs = self._morphs_with_specialchars(self.document)

        # root 생성
        self.tree.clear()
        self.tree.add_node(
            node_id=ParseTree.ID_ROOT,
            node_data=NodeData(
                node_id=ParseTree.ID_ROOT,
                node_type="문서",
                org_txt_form=self.document,
                pos_txt_form=" ".join(
                    [f"{morph[0]}/{morph[1]}" for morph in self.morphs]
                ),
            ),
        )

        self._create_words()

        self._identify_sub_documents()
        self._identify_sentences()

        self._identify_phrases()
        self._identify_clauses()

    def _is_hanja(self, text):
        re_pattern = r"[\u2e80-\u2eff\u31c0-\u31ef\u3200-\u32ff\u3400-\u4dbf\u4e00-\u9fbf\uf900-\ufaff]"
        return re.match(pattern=re_pattern, string=text)

    # mecab이 공백/개행문자등을 걸러내서, 이를 보전하기 위한 유틸리티 함수
    def _morphs_with_specialchars(self, txt):
        old_morphs = self.mecab.pos(txt)
        new_morphs = []
        txt_idx = 0
        for morph, tag in old_morphs:
            morph_idx = 0

            # morph 에 해당하지 않는 문자열이 원문 텍스트에 있는 경우 추출
            start_txt_idx = txt_idx
            while txt[txt_idx] != morph[morph_idx]:
                txt_idx += 1
            end_txt_idx = txt_idx

            # morph 에 해당하는 문자열은 건너뜀
            while morph_idx < len(morph) and txt[txt_idx] == morph[morph_idx]:
                txt_idx += 1
                morph_idx += 1

            # 추출된 신규 형태소 추가
            if start_txt_idx < end_txt_idx:
                new_morphs.append(
                    (txt[start_txt_idx:end_txt_idx], "SWS")
                )  # 특수문자는 SWS 태그를 준다

            # 원래 morph도 신규로 추가
            new_morphs.append((morph, tag))

        return new_morphs

    ##### create_word 관련 함수 시작
    def _create_words(self):
        """Create word nodes from morphs"""
        words = self._words_from_morphs(self.morphs)

        for idx, word in enumerate(words):
            # org_txt, pos_txt 계산
            org_txt_form = ""
            pos_txt_form = ""
            for morph in word:
                org_txt_form += morph[0]
                pos_txt_form += morph[1] if len(pos_txt_form) == 0 else f"+{morph[1]}"
            pos_txt_form = f"{org_txt_form}/{pos_txt_form}"

            self.tree.add_node(
                node_id=f"{ParseTree.ID_ROOT}_{idx:03d}",
                node_data=NodeData(
                    node_id=f"{ParseTree.ID_ROOT}_{idx:03d}",
                    node_type="단어",
                    parent_node_id=ParseTree.ID_ROOT,
                    org_txt_form=org_txt_form,
                    pos_txt_form=pos_txt_form,
                ),
            )

        self._create_composite_words()
        self._create_josa_suffix_words()

    # 형태소 분석결과에서 구분자 태그를 이용해 단어를 추출한다.
    def _words_from_morphs(self, morphs):
        words = []
        word = []

        is_hashtag_mode = False
        is_single_quote_mode = False

        for morph in morphs:
            # hashtag 처리
            if not is_hashtag_mode and len(word) == 0 and morph[0] == "#":
                is_hashtag_mode = True

            if is_hashtag_mode:
                if morph[1] == "SWS":
                    if len(word) > 0:
                        txt, pos = map(list, zip(*word))
                        words.append([("".join(txt), "NNP")])
                        word = []
                    words.append([morph])  # 현재 morph는 delimiter 인데, 독립적인 단어로 만든다
                    is_hashtag_mode = False
                else:
                    if morph[0] == "#" and len(word) > 0:
                        txt, pos = map(list, zip(*word))
                        words.append([("".join(txt), "NNP")])
                        word = []

                    word.append(morph)

                continue

            # ''로 둘러싸인 짧은 문자열(10 형태소 이내)는 하나의 고유명사로 분류한다
            if not is_single_quote_mode and len(word) == 0 and morph[0] == "'":
                words.append([morph])
                is_single_quote_mode = True
                continue

            if is_single_quote_mode:
                if morph[0] == "'":
                    if len(word) > 0:
                        txt, pos = map(list, zip(*word))
                        words.append([("".join(txt), "NNP")])
                        word = []
                    words.append([morph])
                    is_single_quote_mode = False
                else:
                    word.append(morph)
                    if len(word) > 10:
                        for m in word:
                            words.append([m])
                        word = []
                        is_single_quote_mode = False
                continue

            # 외국어가 모두 대문자이면 고유명사로 처리한다
            if morph[1] == "SL" and morph[0].isupper() and len(morph[0]) > 1:
                morph = (morph[0], "NNP")

            # 일반적인 단어 처리
            words.append([morph])

        # for 문 이후 남은 형태소에 대해 word 처리
        if len(word) > 0:
            if is_hashtag_mode:
                txt, pos = map(list, zip(*word))
                words.append([("".join(txt), "NNP")])
            else:
                words.append(word)

        return words

    def _create_composite_words(self):
        delimiters = [
            "SF",
            "SE",
            "SSO",
            "SSC",
            "SC",
            "SY",
            "SWS",  # 기호/공백 문자로 구분
            "JKS",
            "JKC",
            "JKG",
            "JKO",
            "JKB",
            "JKV",
            "JKQ",
            "JX",
            "JC",  # 명사 추출을 위해 관계언을 살려둔다
            "IC",
            "UNKNOWN",  # 감탄사 활용
        ]
        node_ids = self.tree.get_children_node_ids(parent_node_id=ParseTree.ID_ROOT)

        idx = 0
        sub_nodes = []
        while idx < len(node_ids):
            node_data = self.tree.get_node_data_by_id(node_ids[idx])

            if node_data.get_last_pos_tag() in delimiters:
                if len(sub_nodes) > 1:
                    self._create_sub_tree(
                        parent_node_id=ParseTree.ID_ROOT,
                        children_node_data=sub_nodes,
                        node_type="단어",
                    )
                sub_nodes = []
                idx += 1
                continue

            sub_nodes.append(node_data)
            idx += 1

        if len(sub_nodes) > 1:
            self._create_sub_tree(
                parent_node_id=ParseTree.ID_ROOT,
                children_node_data=sub_nodes,
                node_type="단어",
            )

    def _create_josa_suffix_words(self):
        children_ids = self.tree.get_children_node_ids(parent_node_id=ParseTree.ID_ROOT)

        sub_nodes = []
        for idx, child_id in enumerate(children_ids):
            child_data = self.tree.get_node_data_by_id(children_ids[idx])
            sub_nodes.append(child_data)

            if child_data.get_last_pos_tag() == "SWS":
                sub_nodes = []
                continue
            if (
                child_data.get_last_pos_tag() == "SY"
                and len(child_data.org_txt_form) > 1
            ):
                if len(sub_nodes) > 0:
                    self._create_sub_tree(
                        parent_node_id=ParseTree.ID_ROOT,
                        children_node_data=sub_nodes,
                        node_type="단어",
                    )
                sub_nodes = []
                continue
            if child_data.word_tag == "관계언":
                self._create_sub_tree(
                    parent_node_id=ParseTree.ID_ROOT,
                    children_node_data=sub_nodes,
                    node_type="단어",
                )
                sub_nodes = []
                continue

    ##### identify_sub_document 관련 함수 시작
    def _identify_sub_documents(self):
        children_node_ids = self.tree.get_children_node_ids(
            parent_node_id=ParseTree.ID_ROOT
        )

        sub_nodes = []
        is_sub_document = False
        for idx, child_node_id in enumerate(children_node_ids):
            child_node_data = self.tree.get_node_data_by_id(children_node_ids[idx])

            if not is_sub_document and 1 in map(
                lambda x: 1 if child_node_data.org_txt_form.find(x) > -1 else 0,
                ['"', "“"],
            ):
                is_sub_document = True
                sub_nodes.append(child_node_data)
                continue

            if is_sub_document and 1 in map(
                lambda x: 1 if child_node_data.org_txt_form.find(x) > -1 else 0,
                ['"', "”"],
            ):
                sub_nodes.append(child_node_data)
                self._create_sub_tree(
                    parent_node_id=ParseTree.ID_ROOT,
                    children_node_data=sub_nodes,
                    node_type="문서",
                )
                sub_nodes = []
                is_sub_document = False
                continue

            if is_sub_document:
                sub_nodes.append(child_node_data)

    def _identify_sentences(self):
        def is_document_node_processed(tree, node_id):
            return tree.get_node_data_by_id(node_id).node_type == "문장"

        def get_not_processed_document_node_id(tree):
            for node_id in tree.filter_nodes(
                lambda x: tree.get_node_data_by_id(x).node_type == "문서"
            ):
                if 1 not in map(
                    lambda x: 1 if is_document_node_processed(tree, x) else 0,
                    tree.get_children_node_ids(parent_node_id=node_id),
                ):
                    return node_id
            return None

        document_node_id = get_not_processed_document_node_id(self.tree)
        while document_node_id:
            children_node_ids = self.tree.get_children_node_ids(
                parent_node_id=document_node_id
            )

            sub_nodes = []
            idx = 0
            while idx < len(children_node_ids):
                child_node_data = self.tree.get_node_data_by_id(children_node_ids[idx])

                # 문장의 처음에 공백문자를 추가하지 않는다
                if len(sub_nodes) == 0 and child_node_data.get_last_pos_tag() == "SWS":
                    idx += 1
                    continue

                sub_nodes.append(child_node_data)

                if child_node_data.node_type != "단어":
                    idx += 1
                    continue

                # 종결어미, 문장부호(.!?)로 문장구분
                if child_node_data.get_last_pos_tag() in ["EF", "SF"]:
                    idx += 1
                    while idx < len(children_node_ids):
                        next_child_node_data = self.tree.get_node_data_by_id(
                            children_node_ids[idx]
                        )
                        if next_child_node_data.get_last_pos_tag() in [
                            "SF",
                            "SE",
                            "SY",
                            "SWS",
                        ]:
                            sub_nodes.append(next_child_node_data)
                        else:
                            if sub_nodes[-1].get_last_pos_tag() in ["SWS"]:
                                sub_nodes = sub_nodes[0:-1]  # 공백으로 끝나지 않도록 조정한다
                            self._create_sub_tree(
                                parent_node_id=document_node_id,
                                children_node_data=sub_nodes,
                                node_type="문장",
                            )
                            sub_nodes = []
                            sub_nodes.append(next_child_node_data)
                            break

                        idx += 1

                # 종결부호 [.!?]/SY 후 띄어쓰기나 엔터가 오는 문장 처리
                elif (
                    child_node_data.org_txt_form[-1] in [".", "!", "?"]
                    and len(child_node_data.org_txt_form) == 1
                    and child_node_data.get_last_pos_tag() in ["SY"]
                ):
                    next_idx = idx + 1
                    if next_idx < len(children_node_ids):
                        next_child_node_data = self.tree.get_node_data_by_id(
                            children_node_ids[next_idx]
                        )
                        if next_child_node_data.get_last_pos_tag() in ["SWS"]:
                            self._create_sub_tree(
                                parent_node_id=document_node_id,
                                children_node_data=sub_nodes,
                                node_type="문장",
                            )
                            sub_nodes = []
                            idx = next_idx

                # 2개이상의 기호가 연속으로 왔을때 문장으로 구분함
                elif (
                    child_node_data.get_last_pos_tag() in ["SY"]
                    and len(child_node_data.org_txt_form) > 1
                ):
                    idx += 1
                    while idx < len(children_node_ids):
                        next_child_node_data = self.tree.get_node_data_by_id(
                            children_node_ids[idx]
                        )
                        if next_child_node_data.get_last_pos_tag() in [
                            "SY",
                            "SF",
                            "SE",
                        ]:
                            sub_nodes.append(next_child_node_data)
                        else:
                            self._create_sub_tree(
                                parent_node_id=document_node_id,
                                children_node_data=sub_nodes,
                                node_type="문장",
                            )
                            sub_nodes = []
                            sub_nodes.append(next_child_node_data)
                            break

                        idx += 1

                # 말줄임표로 구분된 문장 처리
                elif child_node_data.get_last_pos_tag() in ["SE"]:
                    self._create_sub_tree(
                        parent_node_id=document_node_id,
                        children_node_data=sub_nodes,
                        node_type="문장",
                    )
                    sub_nodes = []

                # 감탄사로 구분된 문장 처리
                elif child_node_data.get_last_pos_tag() in ["IC"]:
                    idx += 1
                    while idx < len(children_node_ids):
                        next_child_node_data = self.tree.get_node_data_by_id(
                            children_node_ids[idx]
                        )
                        if next_child_node_data.get_last_pos_tag() in ["IC", "SWS"]:
                            sub_nodes.append(next_child_node_data)
                        else:
                            if sub_nodes[-1].get_last_pos_tag() in ["SWS"]:
                                sub_nodes = sub_nodes[0:-1]  # 공백으로 끝나지 않도록 조정한다
                            self._create_sub_tree(
                                parent_node_id=document_node_id,
                                children_node_data=sub_nodes,
                                node_type="문장",
                            )
                            sub_nodes = []
                            sub_nodes.append(next_child_node_data)
                            break

                        idx += 1

                idx += 1

            if len(sub_nodes) > 0:
                self._create_sub_tree(
                    parent_node_id=document_node_id,
                    children_node_data=sub_nodes,
                    node_type="문장",
                )
            document_node_id = get_not_processed_document_node_id(self.tree)

    def _identify_phrases(self):
        def is_sentence_node_processed(tree, node_id):
            return tree.get_node_data_by_id(node_id).node_type == "구"

        def get_not_processed_sentence_node_id(tree):
            for node_id in tree.filter_nodes(
                lambda x: tree.get_node_data_by_id(x).node_type == "문장"
            ):
                if 1 not in map(
                    lambda x: 1 if is_sentence_node_processed(tree, x) else 0,
                    tree.get_children_node_ids(parent_node_id=node_id),
                ):
                    return node_id
            return None

        sentence_node_id = get_not_processed_sentence_node_id(self.tree)
        while sentence_node_id:
            children_node_ids = self.tree.get_children_node_ids(
                parent_node_id=sentence_node_id
            )

            sub_nodes = []
            idx = 0
            while idx < len(children_node_ids):
                child_node_data = self.tree.get_node_data_by_id(children_node_ids[idx])

                # 구의 처음에 공백문자를 추가하지 않는다
                if len(sub_nodes) == 0 and child_node_data.get_last_pos_tag() == "SWS":
                    idx += 1
                    continue

                sub_nodes.append(child_node_data)

                if child_node_data.node_type != "단어":
                    idx += 1
                    continue

                # 기본적으로 문장의 주성분 (주어, 목적어, 서술어, 보어)일 때 구를 구분한다
                if child_node_data.sentence_tag in ["주어", "목적어", "서술어", "보어"]:

                    # 목적어 뒤에 오는 관형어를 처리한다: '중국을 방문한 대통령'을 '중국을 방문한' '대통령' 으로 나눈다
                    if child_node_data.sentence_tag in ["목적어"]:
                        idx, sub_nodes = self.check_phrase_to_merge_next_word(
                            idx,
                            sub_nodes,
                            children_node_ids,
                            lambda next_node: next_node.sentence_tag in ["관형어"],
                        )

                    # 서술어 뒤어 오는 보조 용언 VX를 처리한다: '논란을 빚고 있다' 를 '논란을' '빚고 있다' 로 나눈다
                    if child_node_data.sentence_tag in ["서술어"]:
                        idx, sub_nodes = self.check_phrase_to_merge_next_word(
                            idx,
                            sub_nodes,
                            children_node_ids,
                            lambda next_node: next_node.get_first_pos_tag() in ["VX"],
                        )

                    # 동사(VV)로 시작하는 보어가 다음에 서술어를 만나면 합친다: '이긴것이 아니다', '위해서도 낫다'
                    if child_node_data.sentence_tag in [
                        "보어"
                    ] and child_node_data.get_first_pos_tag() in ["VV"]:
                        idx, sub_nodes = self.check_phrase_to_merge_next_word(
                            idx,
                            sub_nodes,
                            children_node_ids,
                            lambda next_node: next_node.sentence_tag in ["서술어"],
                        )

                    # 구 노드를 생성한다
                    self._create_sub_tree(
                        parent_node_id=sentence_node_id,
                        children_node_data=sub_nodes,
                        node_type="구",
                    )
                    sub_nodes = []

                idx += 1

            if len(sub_nodes) > 0:
                self._create_sub_tree(
                    parent_node_id=sentence_node_id,
                    children_node_data=sub_nodes,
                    node_type="구",
                )
            sentence_node_id = get_not_processed_sentence_node_id(self.tree)

    def check_phrase_to_merge_next_word(self, idx, sub_nodes, children_node_ids, rule):
        next_idx = idx + 1
        next_sub_nodes = []
        while next_idx < len(children_node_ids):
            next_child_node_data = self.tree.get_node_data_by_id(
                children_node_ids[next_idx]
            )
            if next_child_node_data.get_last_pos_tag() in ["SWS"]:  # 공백문자를 건너뛴다
                next_sub_nodes.append(next_child_node_data)
                next_idx += 1
                continue
            else:
                break

        if next_idx < len(children_node_ids):
            next_child_node_data = self.tree.get_node_data_by_id(
                children_node_ids[next_idx]
            )
            if rule(next_child_node_data):
                sub_nodes += next_sub_nodes
                sub_nodes.append(next_child_node_data)
                idx = next_idx

        return idx, sub_nodes

    def _identify_clauses(self):
        pass

    ##### 유틸리티 함수 - 트리 분할 / 합병 관련
    def _update_sub_tree_identifier(
        self, sub_root_node_data, new_node_id, new_parent_node_id, updated_data
    ):
        children_node_ids = self.tree.get_children_node_ids(sub_root_node_data.node_id)

        for idx, child_id in enumerate(children_node_ids):
            child_data = self.tree.get_node_data_by_id(child_id)
            self._update_sub_tree_identifier(
                child_data, f"{new_node_id}_{idx:03d}", new_node_id, updated_data
            )

        self.tree.remove_node(sub_root_node_data.node_id)  # 기존 노드는 삭제
        sub_root_node_data.node_id = new_node_id  # 새로운 노드 ID를 부여
        sub_root_node_data.parent_node_id = new_parent_node_id
        updated_data.append(sub_root_node_data)  # 데이터를 유지하여 추후 다시 추가할 수 있도록 함

    def _create_sub_tree(self, parent_node_id, children_node_data, node_type=None):
        if len(children_node_data) > 0:
            # org_txt, pos_txt 계산
            org_txt_form = "".join([node.org_txt_form for node in children_node_data])
            pos_txt_form = " ".join([node.pos_txt_form for node in children_node_data])

            # 기존 node id 업데이트
            new_node_id = children_node_data[0].node_id
            updated_data = []
            for idx, node_data in enumerate(children_node_data):
                self._update_sub_tree_identifier(
                    node_data, f"{new_node_id}_{idx:03d}", new_node_id, updated_data
                )

            # 신규 node 생성
            self.tree.add_node(
                node_id=new_node_id,
                node_data=NodeData(
                    node_id=new_node_id,
                    node_type=node_type,
                    parent_node_id=parent_node_id,
                    org_txt_form=org_txt_form,
                    pos_txt_form=pos_txt_form,
                ),
            )

            # 기존 nodes 를 신규 node 하위로 이동
            for node_data in updated_data:
                self.tree.add_node(node_id=node_data.node_id, node_data=node_data)

    ##### 명사 추출 관련 함수 시작
    def condition_noun_candidates(self, x):
        node_data = self.tree.get_node_data_by_id(x)
        return (
            (node_data.node_type == "단어" and node_data.word_tag in ["체언", "독립언"])
            and (
                len(node_data.org_txt_form) > 1
                or self._is_hanja(node_data.org_txt_form)
            )
            and node_data.org_txt_form
            not in [
                # 불용어
            ]
        )

    def nouns(self):
        result = []

        # 1단어 체언
        for node_id in self.tree.filter_nodes(self.condition_noun_candidates):
            result.append(self.tree.get_node_data_by_id(node_id).org_txt_form)

        return result

    ##### 문장 분리 관련 함수 시작
    def sentences(self):
        result = []
        for node_id in self.tree.filter_nodes(
            lambda x: self.tree.get_node_data_by_id(x).node_type == "문장"
        ):
            node_data = self.tree.get_node_data_by_id(node_id)
            result.append(node_data.org_txt_form)
        return result

    def printable_tree(self, debug=True):
        return self.tree.printable_subtree(
            sub_root_node_id=ParseTree.ID_ROOT, debug=debug
        )



import os

import setuptools


def get_version():
    basedir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(basedir, "kokex", "__init__.py")) as f:
        for line in f.read().splitlines():
            if line.startswith("__version__"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
        else:
            raise RuntimeError("Unable to find version string.")


def get_description():
    with open("README.md", "r") as f:
        return f.read()


setuptools.setup(
    name="kokex",
    version=get_version(),
    author="jsmyung",
    author_email="jsmyung@datansoft.com",
    description="A Korean Keywords Extractor with Syntactic Analysis",
    long_description=get_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/jsmyung-datansoft/kokex",
    packages=setuptools.find_packages(),
    python_requires=">=3.7",
    install_requires=[
        "konlpy>=0.5.2",
        "networkx>=2.5.1",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
